from unittest import TestCase, mock
import os
from datetime import datetime

from servicer.servicer import Servicer

class ServicerTest(TestCase):
    def setUp(self):
        self.servicer = Servicer(args={}, init=False)

        self.servicer.run = mock.Mock()
        self.servicer.token_interpolator = mock.Mock()
        self.servicer.token_interpolator.interpolate_tokens = mock.Mock()
        self.servicer.dependency_grapher = mock.Mock()

    def AutoMock(self, **attributes):
        automock = mock.Mock(spec=attributes.keys())
        for k, v in attributes.items():
            setattr(automock, k, v)
        return automock

class ServicerClassTest(ServicerTest):
    def test_initialized(self):
        pass

class LoadEnvironmentTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.datetime = self.AutoMock(
            now=mock.Mock(return_value=self.AutoMock(strftime=mock.Mock(return_value='1970-01-01'))),
            utcnow=mock.Mock(return_value='1970-01-01 00:00:00'),
        )
        self.servicer.load_env_file = mock.Mock(return_value=None)

        self.args = {}

    def test_environment_variables_set(self):
        os.environ['PWD'] = 'project_path'
        result = self.servicer.load_environment(self.args)
        self.assertEqual(os.environ['PROJECT_PATH'], 'project_path')
        self.assertEqual(os.environ['BUILD_DATETIME'], '1970-01-01 00:00:00')
        self.assertEqual(os.environ['BUILD_DATE'], '1970-01-01')

    def test_no_env_file_paths(self):
        result = self.servicer.load_environment(self.args)
        self.servicer.load_env_file.assert_not_called()

    def test_multiple_env_file_paths(self):
        self.args['env_file_paths'] = 'path/one:path/two'
        result = self.servicer.load_environment(self.args)
        self.servicer.load_env_file.assert_has_calls([
            mock.call('path/one'),
            mock.call('path/two'),
        ])

class LoadEnvFileTest(ServicerTest):
    def setUp(self):
        super().setUp()

    def test_path_does_not_exist(self):
        with mock.patch('builtins.open', create=True) as mock_open:
            result = self.servicer.load_env_file('fake/path.yaml')
            mock_open.assert_not_called()

    def test_valid_env_file_path(self):
        with mock.patch('yaml.load', create=True) as mock_yaml_load:
            with mock.patch('builtins.open', create=True) as mock_open:
                with mock.patch('os.path.exists', create=True) as mock_os_path_exists:
                    mock_yaml_load.return_value = {'FOO': 'BAR'}
                    mock_os_path_exists.return_value = True

                    result = self.servicer.load_env_file('real/path.yaml')
                    self.assertEqual(os.environ['FOO'], 'BAR')

class GetServiceEnvironmentTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.map_service_environment = mock.Mock(return_value=(None, None))

        self.servicer.config = {}
        self.servicer.config['environment'] = {
            'mappings': [{ 'branch': 'foo', 'environment': 'bar' }],
        }

    def test_custom_mappings(self):
        result = self.servicer.get_service_environment('my-branch')
        self.servicer.map_service_environment.assert_called_with('my-branch', self.servicer.config['environment']['mappings'])
        self.assertEqual(result, None)

    def test_matched_environment(self):
        self.servicer.map_service_environment.return_value = ('my_wacky/environment', { 'branch': 'my_wacky/environment' })

        result = self.servicer.get_service_environment('my-branch')
        self.servicer.map_service_environment.assert_called_with('my-branch', mock.ANY)
        self.assertEqual(result, 'my-wacky-environment')

class MapServiceEnvironmentTest(ServicerTest):
    def test_no_mappings(self):
        result = self.servicer.map_service_environment('my-branch', [])
        self.assertEqual(result, (None, None))

    def test_branch_mapping(self):
        mapping = { 'branch': 'develop' }
        result = self.servicer.map_service_environment('develop', [mapping])
        self.assertEqual(result, ('develop', mapping))

    def test_branch_environment_mapping(self):
        mapping = { 'branch': 'master', 'environment': 'production' }
        result = self.servicer.map_service_environment('master', [mapping])
        self.assertEqual(result, ('production', mapping))

    def test_branch_wildcard_mapping(self):
        mapping = { 'branch': 'env-*' }
        result = self.servicer.map_service_environment('env-my-branch', [mapping])
        self.assertEqual(result, ('env-my-branch', mapping))

    def test_regex_mapping(self):
        mapping = { 'branch': '/.*-my-.*/' }
        result = self.servicer.map_service_environment('env-my-branch', [mapping])
        self.assertEqual(result, ('env-my-branch', mapping))

    def test_no_match(self):
        result = self.servicer.map_service_environment('my-branch', [
            { 'branch': 'master', 'environment': 'production' },
            { 'branch': 'develop' },
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, (None, None))

    def test_no_match_but_starts_with_branch(self):
        result = self.servicer.map_service_environment('master_branch', [
            { 'branch': 'master', 'environment': 'production' },
            { 'branch': 'develop' },
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, (None, None))

    def test_multiple_match(self):
        result = self.servicer.map_service_environment('env-my-branch-qa', [
            { 'branch': 'master', 'environment': 'production' },
            { 'branch': '*-qa', 'environment': 'qa' },
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, ('qa', { 'branch': '*-qa', 'environment': 'qa' }))

class RunServiceStepTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.config = {}

        self.adapter = mock.Mock()
        self.adapter.up = mock.Mock(return_value='service-step results')
        self.Service = mock.Mock(return_value=self.adapter)
        self.module = self.AutoMock(Service=self.Service)

        def mock_load_service_module(service):
            service['module'] = self.module

        self.servicer.load_service_module = mock.Mock(side_effect=mock_load_service_module)

        self.service = {
            'module': self.module,
            'name': 'service_1',
            'steps': {
                'build': {'config': {}},
            }
        }

    def test_runs_an_empty_service_step(self):
        self.service.pop('module')
        self.service['steps']['build'].pop('config')

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_not_called()
        self.servicer.token_interpolator.interpolate_tokens.assert_called_with({}, {})

    def test_runs_a_service_step_with_config_and_no_module(self):
        self.service.pop('module')

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.load_service_module.assert_called_with(self.service)

        self.servicer.run.assert_not_called()
        self.servicer.token_interpolator.interpolate_tokens.assert_called_with(self.service['steps']['build'], self.servicer.config)
        self.assertEqual(self.service['steps']['build']['results'], 'service-step results')

    def test_runs_a_service_step_with_module_and_no_config(self):
        self.service['steps']['build'].pop('config')

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_not_called()
        self.servicer.token_interpolator.interpolate_tokens.assert_called_with({}, {})

    def test_runs_a_service_step_with_module_and_config(self):
        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_not_called()
        self.servicer.token_interpolator.interpolate_tokens.assert_called_with(self.service['steps']['build'], self.servicer.config)

        self.assertTrue('git' not in self.service['steps']['build']['config'])
        self.assertEqual(self.service['steps']['build']['results'], 'service-step results')

    def test_runs_a_service_step_with_git_integration(self):
        self.servicer.config['git'] = {'enabled': True}
        self.servicer.git = {}

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_not_called()
        self.servicer.token_interpolator.interpolate_tokens.assert_called_with(self.service['steps']['build'], self.servicer.config)

        self.assertEqual(self.service['steps']['build']['config']['git'], {'module': {}})
        self.assertEqual(self.service['steps']['build']['results'], 'service-step results')

    def test_runs_a_service_step_with_commands(self):
        self.service['steps']['build']['commands'] = [
            'pre-command1.sh',
            'rm -rf treeeeee xD',
        ]
        self.service['steps']['build']['post_commands'] = [
            'cowsay moo',
            'yes | lolcat',
        ]

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_has_calls([
            mock.call('pre-command1.sh'),
            mock.call('rm -rf treeeeee xD'),
            mock.call('cowsay moo'),
            mock.call('yes | lolcat'),
        ])



from .base_auth_adapter import BaseAuthAdapter

import os
from servicer.run import run

class AuthAdapter(BaseAuthAdapter):
    def __init__(self, config, logger=None):
        super().__init__(config, logger=logger)
        self.run = run

    def authenticate(self):
        self.run('echo "$DOCKER_PASSWORD" | docker login -u "$DOCKER_USERNAME" --password-stdin')



from unittest import TestCase, mock

from servicer.token_interpolator import TokenInterpolator

class TokenInterpolatorTest(TestCase):
    def setUp(self):
        self.token_interpolator = TokenInterpolator()

class TokenInterpolatorClassTest(TokenInterpolatorTest):
    def test_initialized(self):
        pass

class InterpolateTokensTest(TokenInterpolatorTest):
    def setUp(self):
        super().setUp()

        self.token_interpolator.interpolate_tokens = mock.Mock(side_effect=self.token_interpolator.interpolate_tokens)
        self.token_interpolator.replace_tokens = mock.Mock()

    def test_handles_empty_dicts(self):
        config = {}
        self.token_interpolator.interpolate_tokens(config, {})

        self.assertEqual(self.token_interpolator.interpolate_tokens.mock_calls, [
            mock.call(config, {}),
        ])
        self.token_interpolator.replace_tokens.assert_not_called()

    def test_handles_empty_lists(self):
        config = []
        self.token_interpolator.interpolate_tokens(config, {})

        self.assertEqual(self.token_interpolator.interpolate_tokens.mock_calls, [
            mock.call(config, {}),
        ])
        self.token_interpolator.replace_tokens.assert_not_called()

    def test_handles_none(self):
        config = None
        self.token_interpolator.interpolate_tokens(config, {})

        self.assertEqual(self.token_interpolator.interpolate_tokens.mock_calls, [
            mock.call(config, {}),
        ])
        self.token_interpolator.replace_tokens.assert_not_called()

    def test_ignoring_non_string_values(self):
        config = {
            'foo': 123,
        }
        self.token_interpolator.interpolate_tokens(config, {})

        self.assertEqual(self.token_interpolator.interpolate_tokens.mock_calls, [
            mock.call(config, {}),
            mock.call(123, {}, False),
        ])
        self.token_interpolator.replace_tokens.assert_not_called()

    def test_replacing_string_dicts(self):
        config = {
            'foo': 'bar',
        }
        self.token_interpolator.interpolate_tokens(config, {})

        self.assertEqual(self.token_interpolator.interpolate_tokens.mock_calls, [
            mock.call(config, {}),
        ])
        self.assertEqual(self.token_interpolator.replace_tokens.mock_calls, [
            mock.call('bar', {}, False),
        ])

    def test_replacing_string_lists(self):
        config = ['foo', 'bar']
        self.token_interpolator.interpolate_tokens(config, {})

        self.assertEqual(self.token_interpolator.interpolate_tokens.mock_calls, [
            mock.call(config, {}),
        ])
        self.assertEqual(self.token_interpolator.replace_tokens.mock_calls, [
            mock.call('foo', {}, False),
            mock.call('bar', {}, False),
        ])

    def test_replacing_complex_nested_strings(self):
        config = {
            'colors': ['red', 'blue'],
            'fish': {
                'tilapia': 'ok',
                'salmon': 'nice',
                'tuna': {
                    'flavors': ['fishy', 'fresh', 'salty'],
                    'feelings': 'various',
                },
            },
        }
        self.token_interpolator.interpolate_tokens(config, {})

        self.assertEqual(self.token_interpolator.interpolate_tokens.mock_calls, [
            mock.call(config, {}),
            mock.call(config['colors'], {}, False),
            mock.call(config['fish'], {}, False),
            mock.call(config['fish']['tuna'], {}, False),
            mock.call(config['fish']['tuna']['flavors'], {}, False),
        ])
        self.assertEqual(self.token_interpolator.replace_tokens.mock_calls, [
            mock.call('red', {}, False),
            mock.call('blue', {}, False),
            mock.call('ok', {}, False),
            mock.call('nice', {}, False),
            mock.call('fishy', {}, False),
            mock.call('fresh', {}, False),
            mock.call('salty', {}, False),
            mock.call('various', {}, False),
        ])

    def test_passing_params(self):
        config = {
            'colors': ['red', 'blue'],
        }
        self.token_interpolator.interpolate_tokens(config, {'foo': 'bar'})

        self.assertEqual(self.token_interpolator.interpolate_tokens.mock_calls, [
            mock.call(config, {'foo': 'bar'}),
            mock.call(config['colors'], {'foo': 'bar'}, False),
        ])
        self.assertEqual(self.token_interpolator.replace_tokens.mock_calls, [
            mock.call('red', {'foo': 'bar'}, False),
            mock.call('blue', {'foo': 'bar'}, False),
        ])

    def test_passing_ignore_missing_key(self):
        config = {
            'colors': ['red', 'blue'],
        }
        self.token_interpolator.interpolate_tokens(config, {}, ignore_missing_key=True)

        self.assertEqual(self.token_interpolator.interpolate_tokens.mock_calls, [
            mock.call(config, {}, ignore_missing_key=True),
            mock.call(config['colors'], {}, True),
        ])
        self.assertEqual(self.token_interpolator.replace_tokens.mock_calls, [
            mock.call('red', {}, True),
            mock.call('blue', {}, True),
        ])

class ReplaceTokensTest(TokenInterpolatorTest):
    def setUp(self):
        super().setUp()

        self.token_interpolator.evaluate_token = mock.Mock(return_value=None)

    def test_handles_empty_string(self):
        result = self.token_interpolator.replace_tokens('', {})

        self.token_interpolator.evaluate_token.assert_not_called()
        self.assertEqual(result, '')

    def test_does_not_replace_a_string(self):
        result = self.token_interpolator.replace_tokens('foo', {})

        self.token_interpolator.evaluate_token.assert_not_called()
        self.assertEqual(result, 'foo')

    def test_replaces_a_string(self):
        self.token_interpolator.evaluate_token.return_value = 'red'

        result = self.token_interpolator.replace_tokens('${ONE}', {})

        self.assertEqual(self.token_interpolator.evaluate_token.mock_calls, [
            mock.call('${ONE}', {}),
        ])
        self.assertEqual(result, 'red')

    def test_replaces_multiple_strings(self):
        self.token_interpolator.evaluate_token.side_effect = [
            'red',
            'blue',
        ]

        result = self.token_interpolator.replace_tokens('${ONE}+${TWO}', {})

        self.assertEqual(self.token_interpolator.evaluate_token.mock_calls, [
            mock.call('${ONE}', {}),
            mock.call('${TWO}', {}),
        ])
        self.assertEqual(result, 'red+blue')

    def test_replaces_nested_strings(self):
        self.token_interpolator.evaluate_token.side_effect = [
            'red',
        ]

        result = self.token_interpolator.replace_tokens('${ONE:${TWO}}', {})

        self.assertEqual(self.token_interpolator.evaluate_token.mock_calls, [
            mock.call('${ONE:${TWO}}', {}),
        ])
        self.assertEqual(result, 'red')

class EvaluateTokenTest(TokenInterpolatorTest):
    def setUp(self):
        super().setUp()

        # once I think hard enough about how to recurse this
        # self.token_interpolator.evaluate_token = mock.Mock(side_effect=self.token_interpolator.evaluate_token)
        self.token_interpolator.evaluate_value = mock.Mock(return_value=None)

    def test_handles_default_values_with_no_evaluated_value(self):
        result = self.token_interpolator.evaluate_token('${TWO:family.genus.species}', {})

        self.assertEqual(self.token_interpolator.evaluate_value.mock_calls, [
            mock.call('TWO', {}),
            mock.call('family.genus.species', {}),
        ])
        self.assertEqual(result, None)

    def test_handles_default_values_with_an_evaluated_value(self):
        self.token_interpolator.evaluate_value = mock.Mock(side_effect=[
            None,
            'bear',
        ])

        result = self.token_interpolator.evaluate_token('${TWO:family.genus.species}', {})

        self.assertEqual(self.token_interpolator.evaluate_value.mock_calls, [
            mock.call('TWO', {}),
            mock.call('family.genus.species', {}),
        ])
        self.assertEqual(result, 'bear')

    def test_ignores_default_value_with_an_evaluated_value(self):
        self.token_interpolator.evaluate_value = mock.Mock(side_effect=[
            'blue',
        ])

        result = self.token_interpolator.evaluate_token('${TWO:family.genus.species}', {})

        self.assertEqual(self.token_interpolator.evaluate_value.mock_calls, [
            mock.call('TWO', {}),
        ])
        self.assertEqual(result, 'blue')

    def test_finds_single_evaluated_value(self):
        self.token_interpolator.evaluate_value = mock.Mock(side_effect=[
            'blue',
        ])

        result = self.token_interpolator.evaluate_token('${TWO}', {})

        self.assertEqual(self.token_interpolator.evaluate_value.mock_calls, [
            mock.call('TWO', {}),
        ])
        self.assertEqual(result, 'blue')

class EvaluateValueTest(TokenInterpolatorTest):
    def setUp(self):
        super().setUp()

        self.token_interpolator.dict_get_path = mock.Mock(return_value=None)

    def test_evaluates_a_single_string(self):
        result = self.token_interpolator.evaluate_value('tacos', {})

        self.token_interpolator.dict_get_path.assert_called_with(
            path=['tacos'],
            _dict={},
            ignore_missing_key=True,
        )
        self.assertEqual(result, None)

    def test_evaluates_a_string_literal(self):
        result = self.token_interpolator.evaluate_value('"tacos"', {})

        self.token_interpolator.dict_get_path.assert_not_called()
        self.assertEqual(result, 'tacos')

    def test_evaluates_a_string_path(self):
        params = {
            'crunchy': {
                'supreme': {
                    'doritos': 'locos',
                },
            },
        }
        self.token_interpolator.dict_get_path.return_value = 'locos'
        result = self.token_interpolator.evaluate_value('crunchy.supreme.doritos', params)

        self.token_interpolator.dict_get_path.assert_called_with(
            path=['crunchy', 'supreme', 'doritos'],
            _dict=params,
            ignore_missing_key=True,
        )
        self.assertEqual(result, 'locos')

    def test_evaluates_string_but_not_found(self):
        result = self.token_interpolator.evaluate_value('crunchy.supreme.doritos', {})

        self.token_interpolator.dict_get_path.assert_called_with(
            path=['crunchy', 'supreme', 'doritos'],
            _dict={},
            ignore_missing_key=True,
        )
        self.assertEqual(result, None)

class DictGetPathTest(TokenInterpolatorTest):
    def setUp(self):
        super().setUp()

        self.token_interpolator.dict_get_path = mock.Mock(side_effect=self.token_interpolator.dict_get_path)

    def test_none_path(self):
        result = self.token_interpolator.dict_get_path(path=None, _dict={})

        self.assertEqual(self.token_interpolator.dict_get_path.mock_calls, [
            mock.call(path=None, _dict={}),
        ])
        self.assertEqual(result, None)

    def test_empty_path(self):
        result = self.token_interpolator.dict_get_path(path=[], _dict={})

        self.assertEqual(self.token_interpolator.dict_get_path.mock_calls, [
            mock.call(path=[], _dict={}),
        ])
        self.assertEqual(result, None)

    def test_simple_path(self):
        path = ['egg']
        _dict = {'egg': 'whites'}
        result = self.token_interpolator.dict_get_path(path=path, _dict=_dict)

        self.assertEqual(self.token_interpolator.dict_get_path.mock_calls, [
            mock.call(path=path, _dict=_dict),
        ])
        self.assertEqual(result, 'whites')

    def test_a_magnificent_path(self):
        path = ['terrifying', 'beautiful', 'powerful']
        _dict = {
            'terrifying': {
                'beautiful': {
                    'powerful': 'grey prince zote',
                },
            },
        }
        result = self.token_interpolator.dict_get_path(path=path, _dict=_dict)

        self.assertEqual(self.token_interpolator.dict_get_path.mock_calls, [
            mock.call(path=['terrifying', 'beautiful', 'powerful'], _dict=_dict),
            mock.call(path=['beautiful', 'powerful'], _dict=_dict['terrifying'], ignore_missing_key=False),
            mock.call(path=['powerful'], _dict=_dict['terrifying']['beautiful'], ignore_missing_key=False),
        ])
        self.assertEqual(result, 'grey prince zote')

    def test_throws_key_error_for_invalid_path(self):
        path = ['terrifying', 'awful', 'powerful']
        _dict = {
            'terrifying': {
                'beautiful': {
                    'powerful': 'grey prince zote',
                },
            },
        }
        with self.assertRaises(KeyError) as context:
            result = self.token_interpolator.dict_get_path(path=path, _dict=_dict)

        self.assertTrue('awful' in str(context.exception))

    def test_allows_missing_path(self):
        path = ['terrifying', 'awful', 'powerful']
        _dict = {
            'terrifying': {
                'beautiful': {
                    'powerful': 'grey prince zote',
                },
            },
        }
        result = self.token_interpolator.dict_get_path(path=path, _dict=_dict, ignore_missing_key=True)

        self.assertEqual(self.token_interpolator.dict_get_path.mock_calls, [
            mock.call(path=['terrifying', 'awful', 'powerful'], _dict=_dict, ignore_missing_key=True),
            mock.call(path=['awful', 'powerful'], _dict=_dict['terrifying'], ignore_missing_key=True),
        ])
        self.assertEqual(result, None)



from unittest import TestCase, mock

from servicer.dependency_grapher import DependencyGrapher

class DependencyGrapherTest(TestCase):
    def setUp(self):
        self.logger = mock.Mock()
        self.dependency_grapher = DependencyGrapher(
            config={
                'args': {'ignore_dependencies': False},
                'graph': {'implicit-step-dependencies': True},
                'services': {
                    'service_1': {
                        'name': 'service_1',
                        'steps': {'build': {}, 'test': {}},
                    },
                    'service_2': {
                        'name': 'service_2',
                        'steps': {'build': {}, 'test': {}, 'deploy': {}},
                    },
                },
            },
            active_services=[
                'service_1',
                'service_2',
            ],
            steps={
                'build': {'config': {}},
                'test': {'config': {}},
                'deploy': {'config': {}},
            },
            step_order=[
                'build',
                'test',
                'deploy',
            ],
            active_steps=[
                'build',
                'test',
                'deploy',
            ],
            logger=self.logger,
        )
        self.dependency_grapher.toposort2 = mock.Mock()

class DependencyGrapherTest(DependencyGrapherTest):
    def test_initialized(self):
        pass

class OrderServiceStepsTest(DependencyGrapherTest):
    def setUp(self):
        super().setUp()

        self.dependency_grapher.add_dependencies = mock.Mock()
        self.dependency_grapher.toposort2 = mock.Mock(return_value=[
            ['service_1:build', 'service_2:build'],
            ['service_1:test', 'service_2:test'],
        ])

    def test_returns_a_service_step_order(self):
        result = self.dependency_grapher.order_service_steps(self.dependency_grapher.config['services'].keys())

        self.assertEqual(self.dependency_grapher.add_dependencies.mock_calls, [
            mock.call({}, 'service_1:build', True),
            mock.call({}, 'service_1:test', True),
            mock.call({}, 'service_2:build', True),
            mock.call({}, 'service_2:test', True),
            mock.call({}, 'service_2:deploy', True),
        ])
        self.dependency_grapher.toposort2.assert_called_with({})
        self.assertEqual(result, [
            ['service_1:build', 'service_2:build'],
            ['service_1:test', 'service_2:test'],
        ])

    def test_returns_an_empty_service_step_order(self):
        self.dependency_grapher.toposort2 = mock.Mock(return_value=[])

        result = self.dependency_grapher.order_service_steps([])

        self.dependency_grapher.toposort2.assert_called_with({})
        self.assertEqual(result, [])

    def test_removes_orphaned_dependencies(self):
        def mock_add_dependencies(dependencies, service_step_name, follow_dependencies=True):
            dependencies['service_1:test'] = set(['service_1:build'])

        self.dependency_grapher.add_dependencies = mock.Mock(side_effect=mock_add_dependencies)
        self.dependency_grapher.config['args']['ignore_dependencies'] = True

        result = self.dependency_grapher.order_service_steps(self.dependency_grapher.config['services'].keys())

        self.assertEqual(self.dependency_grapher.add_dependencies.mock_calls, [
            mock.call({'service_1:test': set()}, 'service_1:build', False),
            mock.call({'service_1:test': set()}, 'service_1:test', False),
            mock.call({'service_1:test': set()}, 'service_2:build', False),
            mock.call({'service_1:test': set()}, 'service_2:test', False),
            mock.call({'service_1:test': set()}, 'service_2:deploy', False),
        ])
        self.dependency_grapher.toposort2.assert_called_with({'service_1:test': set()})
        self.assertEqual(result, [
            ['service_1:build', 'service_2:build'],
            ['service_1:test', 'service_2:test'],
        ])

class ServiceStepDependsOnTest(DependencyGrapherTest):
    def setUp(self):
        super().setUp()

        self.dependency_grapher.get_depends_on = mock.Mock(return_value=[])
        self.dependencies = {}

    def test_empty_depends_on(self):
        result = self.dependency_grapher.service_step_depends_on('service_1:build')

        self.assertEqual(self.dependency_grapher.get_depends_on.mock_calls, [
            mock.call({'name': 'service_1', 'steps': {'build': {}, 'test': {}}}),
            mock.call({}),
        ])
        self.assertEqual(result, [])

    def test_depends_on_from_service_config(self):
        self.dependency_grapher.get_depends_on.side_effect = [
            ['service_2:build', 'service_3:build'],
            [],
        ]

        result = self.dependency_grapher.service_step_depends_on('service_1:build')

        self.assertEqual(result, ['service_2:build', 'service_3:build'])

    def test_depends_on_from_service_step_config(self):
        self.dependency_grapher.get_depends_on.side_effect = [
            [],
            ['service_2:test', 'service_3:test'],
        ]

        result = self.dependency_grapher.service_step_depends_on('service_1:build')

        self.assertEqual(result, ['service_2:test', 'service_3:test'])

    def test_depends_on_from_implicit_dependencies(self):
        result = self.dependency_grapher.service_step_depends_on('service_1:test')

        self.assertEqual(result, ['service_1:build'])

    def test_depends_on_with_no_implicit_dependencies(self):
        self.dependency_grapher.config['graph']['implicit-step-dependencies'] = False

        result = self.dependency_grapher.service_step_depends_on('service_1:test')

        self.assertEqual(result, [])

    def test_depends_on_chooses_automatic_step(self):
        self.dependency_grapher.config['services']['service_3'] = {}

        self.dependency_grapher.get_depends_on.side_effect = [
            ['service_2'],
            ['service_3'],
        ]

        result = self.dependency_grapher.service_step_depends_on('service_1:build')

        self.assertEqual(result, ['service_2:build'])

class GetDependsOnTest(DependencyGrapherTest):
    def setUp(self):
        super().setUp()

    def test_returns_empty_list_if_no_depends_on(self):
        result = self.dependency_grapher.get_depends_on({})
        self.assertEqual(result, [])

    def test_returns_list_with_depends_on_as_list(self):
        result = self.dependency_grapher.get_depends_on({'depends_on': ['blackberries', 'blueberries']})
        self.assertEqual(result, ['blackberries', 'blueberries'])

    def test_returns_list_with_depends_on_as_value(self):
        result = self.dependency_grapher.get_depends_on({'depends_on': 'raspberries'})
        self.assertEqual(result, ['raspberries'])

class AddDependenciesTest(DependencyGrapherTest):
    def setUp(self):
        super().setUp()

        self.dependency_grapher.service_step_depends_on = mock.Mock(return_value=[])
        self.dependency_grapher.add_dependency = mock.Mock()

        self.dependency_grapher.config['services'] = {
            'service_1': {
                'name': 'service_1',
                'steps': {'build': {}, 'test': {}},
            },
            'service_2': {
                'name': 'service_2',
                'steps': {'build': {}, 'test': {}, 'deploy': {}},
            },
            'service_3': {
                'name': 'service_3',
                'steps': {'build': {}, 'test': {}},
            },
            'service_4': {
                'name': 'service_4',
                'steps': {'build': {}, 'deploy': {}},
            },
        }

        self.dependency_grapher.active_services = ['service_1', 'service_2']
        self.dependencies = {}

    def test_skips_if_an_entry_already_exists(self):
        self.dependencies = {'service_1:build': set(['strawberry', 'mango', 'banana'])}

        result = self.dependency_grapher.add_dependencies(self.dependencies, 'service_1:build')

        self.dependency_grapher.service_step_depends_on.assert_not_called()
        self.dependency_grapher.add_dependency.assert_not_called()
        self.assertEqual(self.dependencies, {
            'service_1:build': set(['strawberry', 'mango', 'banana']),
        })

    def test_adds_an_empty_set_with_no_dependencies(self):
        result = self.dependency_grapher.add_dependencies(self.dependencies, 'service_1:build')

        self.assertEqual(self.dependency_grapher.service_step_depends_on.mock_calls, [
            mock.call('service_1:build'),
        ])
        self.dependency_grapher.add_dependency.assert_not_called()
        self.assertEqual(self.dependencies, {
            'service_1:build': set(),
        })

    def test_adds_a_service_step_with_dependencies(self):
        self.dependency_grapher.service_step_depends_on.return_value = [
            'service_2:build',
            'service_3:build',
        ]

        result = self.dependency_grapher.add_dependencies(self.dependencies, 'service_1:build')

        self.assertEqual(self.dependency_grapher.add_dependency.mock_calls, [
            mock.call({'service_1:build': set()}, 'service_1:build', 'service_2', 'build', True),
            mock.call({'service_1:build': set()}, 'service_1:build', 'service_3', 'build', True),
        ])

    def test_adds_a_soft_service_wildcard_dependency(self):
        self.dependency_grapher.service_step_depends_on.return_value = [
            '*:build',
        ]

        result = self.dependency_grapher.add_dependencies(self.dependencies, 'service_1:build')

        self.assertEqual(self.dependency_grapher.add_dependency.mock_calls, [
            mock.call({'service_1:build': set()}, 'service_1:build', '*', 'build', False),
        ])

    def test_adds_a_hard_service_wildcard_dependency(self):
        self.dependency_grapher.service_step_depends_on.return_value = [
            '*:build!',
        ]

        result = self.dependency_grapher.add_dependencies(self.dependencies, 'service_1:build')

        self.assertEqual(self.dependency_grapher.add_dependency.mock_calls, [
            mock.call({'service_1:build': set()}, 'service_1:build', '*', 'build!', False),
        ])

    def test_throws_an_error_for_an_invalid_service_dependency(self):
        self.dependency_grapher.service_step_depends_on.return_value = [
            'service_9000:build',
        ]

        with self.assertRaises(ValueError) as context:
            result = self.dependency_grapher.add_dependencies(self.dependencies, 'service_1:build')

        self.dependency_grapher.add_dependency.assert_not_called()
        self.assertTrue('Invalid service dependency specified: service_9000:build, "service_9000" must be included in services: [service_1,service_2,service_3,service_4]' in str(context.exception))

    def test_throws_an_error_for_an_invalid_step_dependency(self):
        self.dependency_grapher.service_step_depends_on.return_value = [
            'service_2:party',
        ]

        with self.assertRaises(ValueError) as context:
            result = self.dependency_grapher.add_dependencies(self.dependencies, 'service_1:build')

        self.dependency_grapher.add_dependency.assert_not_called()
        self.assertTrue('Invalid step dependency specified: service_2:party, "party" must be included in steps: [build,test,deploy]' in str(context.exception))

class AddDependencyTest(DependencyGrapherTest):
    def setUp(self):
        super().setUp()

        self.dependency_grapher.add_dependencies = mock.Mock()

        self.dependency_grapher.active_services = ['service_1', 'service_2']
        self.dependencies = {'service_1:build': set()}

    def test_adds_a_dependency_without_following(self):
        result = self.dependency_grapher.add_dependency(self.dependencies, 'service_1:build', 'service_2', 'build', False)

        self.dependency_grapher.add_dependencies.assert_not_called()
        self.assertEqual(self.dependencies, {
            'service_1:build': set(['service_2:build']),
        })

    def test_adds_a_dependency_with_following(self):
        result = self.dependency_grapher.add_dependency(self.dependencies, 'service_1:build', 'service_2', 'build', True)

        self.dependency_grapher.add_dependencies.assert_called_with({'service_1:build': {'service_2:build'}}, 'service_2:build', True)
        self.assertEqual(self.dependencies, {
            'service_1:build': set(['service_2:build']),
        })



from .base_auth_adapter import BaseAuthAdapter

import urllib.request
import os
from servicer.run import run

class AuthAdapter(BaseAuthAdapter):
    def __init__(self, config, logger=None):
        super().__init__(config, logger=logger)
        self.run = run

        self.sonarqube_token = os.getenv('SONARQUBE_TOKEN', self.config.get('auth_token'))

        self.scanner_version = self.config['scanner_version']
        self.scanner_path = self.config.get('scanner_path', './')

        if self.scanner_path.endswith('/'):
            self.scanner_path = '%ssonar-scanner-cli-%s' % (self.scanner_path, self.scanner_version)

    def authenticate(self):
        if self.sonarqube_token:
            self.ensure_install()
        else:
            self.logger.log('no sonarqube token found, skipping')

    def ensure_install(self):
        if os.path.exists(self.scanner_path):
            self.logger.log('found sonar-scanner installation at %s' % self.scanner_path)
        else:
            download_file = 'sonar-scanner-cli-%s.zip' % self.scanner_version
            download_directory = os.path.dirname(self.scanner_path)
            download_path = '%s/%s' % (download_directory, download_file)
            url = 'https://binaries.sonarsource.com/Distribution/sonar-scanner-cli/%s' % download_file

            file_path = '%s.zip' % self.scanner_path
            self.logger.log('downloading sonar-scanner from %s to %s' % (url, download_path))
            urllib.request.urlretrieve(url, download_path)

            self.run('unzip -d `dirname %s` %s' % (download_path, download_path))
            self.run('rm %s' % download_path)

            extracted_directory = 'sonar-scanner-%s' % self.scanner_version
            self.run('mv %s/%s %s' % (download_directory, extracted_directory, self.scanner_path))



import os
import requests
import json

from ..task_service import Service as BaseService
from ..request_service import Service as RequestService

class Service(BaseService, RequestService):
    def __init__(self, config, logger=None):
        super().__init__(config, logger=logger)

        self.base_url = self.config.get('endpoint', os.environ['CONSUL_ENDPOINT'])
        self.token = self.config.get('token', os.environ['CONSUL_TOKEN'])
        self.base_path = self.config.get('base_path', '')

    def write_file(self, files=[]):
        if not isinstance(files, list):
            files = [files]

        for f in files:
            with open(f['file_path'], 'r') as _file:
                data = _file.read()

                data = self.token_interpolator.replace_tokens(data, os.environ, ignore_missing_key=True)

                consul_url = '%s%s' % (self.base_path, f['consul_path'])
                self.logger.log('writing file to consul: %s -> %s' % (f['file_path'], consul_url))

                self.consul_request(
                    data=data,
                    method='put',
                    url=consul_url,
                )

    def read(self, **params):
        response = self.consul_request(url='%s%s' % (self.base_path, params['path']))
        self.logger.log(response.text)

    def consul_request(self, **params):
        if 'headers' not in params:
            params['headers'] = {}
        params['headers']['X-Consul-Token'] = self.token

        return self.request(**params)



import json
import os

from .topological_order import toposort2

class DependencyGrapher():
    def __init__(self, config, active_services, steps, step_order, active_steps, logger=None):
        self.toposort2 = toposort2
        self.logger = logger
        self.config = config
        self.active_services = active_services
        self.steps = steps
        self.step_order = step_order
        self.active_steps = active_steps

    def order_service_steps(self, services):
        follow_dependencies = True
        if self.config['args']['ignore_dependencies']:
            follow_dependencies = False

        dependencies = {}

        for service_name in services:
            service = self.config['services'][service_name]

            for step_name in self.active_steps:
                if step_name in service['steps']:
                    service_step_name = '%s:%s' % (service_name, step_name)
                    self.add_dependencies(dependencies, service_step_name, follow_dependencies)

        self.add_wildcard_dependencies(dependencies, follow_dependencies)

        # remove orphaned build dependencies
        if self.config['args']['ignore_dependencies']:
            dep_keys = dependencies.keys()
            for key, value in dependencies.items():
                dependencies[key] = set([v for v in value if v in dep_keys])

        self.logger.log('Dependency Graph:')
        self.logger.log(json.dumps(dependencies, indent=4, sort_keys=True, default=str))
        return self.toposort2(dependencies)

    def service_step_depends_on(self, service_step_name):
        name_pieces = service_step_name.split(':')
        service_name = name_pieces[0]
        service = self.config['services'][service_name]
        step_name = name_pieces[1]

        depends_on = []

        # service level user-defined custom dependencies
        depends_on.extend(self.get_depends_on(service))

        # step level user-defined custom dependencies
        depends_on.extend(self.get_depends_on(service['steps'][step_name]))

        # default step based dependencies
        if self.config['graph']['implicit-step-dependencies']:
            step_index = self.step_order.index(step_name)
            if step_index > 0:
                dep_step_name = self.step_order[step_index - 1]
                if dep_step_name in service['steps']:
                    depends_on.append('%s:%s' % (service_name, self.step_order[step_index - 1]))

        # append implied step dependencies
        depends_on = self.fill_implied_step_dependencies(depends_on, step_name)

        return depends_on

    def fill_implied_step_dependencies(self, depends_on, step_name):
        deps = []
        for d in depends_on:
            if ':' in d:
                deps.append(d)
            elif 'steps' in self.config['services'][d] and step_name in self.config['services'][d]['steps']:
                deps.append('%s:%s' % (d, step_name))
        return deps

    def get_depends_on(self, config):
        depends_on = []
        if 'depends_on' in config:
            if isinstance(config['depends_on'], list):
                depends_on.extend(config['depends_on'])
            else:
                depends_on.append(config['depends_on'])
        return depends_on

    def add_dependencies(self, dependencies, service_step_name, follow_dependencies=True):
        if service_step_name in dependencies:
            return  # only calculate dependencies for each service-step once
        else:
            dependencies[service_step_name] = set()

        depends_on = self.service_step_depends_on(service_step_name)

        service_step_name_pieces = service_step_name

        for dep in depends_on:
            dep_pieces = dep.split(':')
            service_dependency = dep_pieces[0]
            step_dependency = dep_pieces[1]

            if service_dependency == '*':
                # self.delayed_dependencies.append(service_step_name)
                self.add_dependency(dependencies, service_step_name, service_dependency, step_dependency, False)
                continue
            elif service_dependency not in self.config['services']:
                msg = 'Invalid service dependency specified: %s, "%s" must be included in services: [%s]' % (dep, service_dependency, ','.join(self.config['services'].keys()))
                raise ValueError(msg)

            if step_dependency in self.steps:
                self.add_dependency(dependencies, service_step_name, service_dependency, step_dependency, follow_dependencies)
            else:
                msg = 'Invalid step dependency specified: %s, "%s" must be included in steps: [%s]' % (dep, step_dependency, ','.join(self.steps.keys()))
                raise ValueError(msg)

    def add_dependency(self, dependencies, service_step_name, service_name, step_name, follow_dependencies):
        dep_service_step_name = '%s:%s' % (service_name, step_name)
        dependencies[service_step_name].add(dep_service_step_name)
        if follow_dependencies:
            self.add_dependencies(dependencies, dep_service_step_name, follow_dependencies)

    def add_wildcard_dependencies(self, dependencies, follow_dependencies=True):
        dependency_list = list(dependencies)
        for service_step_name in reversed(dependency_list):
            depends_on = dependencies[service_step_name].copy()
            for dep in depends_on:
                if dep.startswith('*'):
                    dependencies[service_step_name].remove(dep)
                    hard_dependency = False

                    if dep.endswith('!'):
                        dep = dep[0:-1]
                        hard_dependency = True

                    step_dependency = dep.split(':')[1]

                    all_services = []
                    if hard_dependency:
                        all_services = self.config['services'].keys()
                    else:
                        service_names = [d.split(':')[0] for d in dependencies.keys()]
                        all_services = list(set(service_names))

                    for _service_name in all_services:
                        if step_dependency in self.config['services'][_service_name]['steps']:
                            self.add_dependency(dependencies, service_step_name, _service_name, step_dependency, follow_dependencies)
